from .models import user_seen


class LastActiveMiddleware(object):
    """
    Middlewate to set timestampe when a user
    has been last seen
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):

        if request.user.is_authenticated:
            user_seen(request.user)

        return self.get_response(request)



from django.contrib import admin

# Register your models here.
from .models import LastActive


class LastActiveAdmin(admin.ModelAdmin):
    list_filter = ("site", "module", "last_active")
    search_fields = ("user__username",)
    list_display = ("site", "module", "user", "last_active")


admin.site.register(LastActive, LastActiveAdmin)



# Create your tests here.
import datetime
import time

import mock
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.core.cache import cache
from django.test import TestCase
from django.utils import timezone
from last_active import middleware, settings
from last_active.models import LastActive, clear_interval, user_seen


class TestLastActiveModel(TestCase):
    def test_unicode(self):
        user = User(username="testuser")
        ts = datetime.datetime(2021, 1, 1, 2, 3, 4)
        seen = LastActive(user=user, last_active=ts)
        self.assertIn("testuser", str(seen))
        self.assertIn("2013-01-01 02:03:04", str(seen))


class TestLastActiveManager(TestCase):
    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen(self, get_or_create):
        user = User(username="testuser", pk=1)
        LastActive = mock.Mock(LastActive)
        get_or_create.return_value = (LastActive, True)

        LastActive.objects.seen(user=user)

        get_or_create.assert_called_with(
            user=user, module=settings.last_active_DEFAULT_MODULE, site=Site.objects.get_current()
        )
        self.assertFalse(LastActive.save.called)

    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen_no_default(self, get_or_create):
        user = User(username="testuser", pk=1)
        site = Site(pk=2)
        get_or_create.return_value = (None, True)

        LastActive.objects.seen(user=user, site=site, module="test")

        get_or_create.assert_called_with(user=user, module="test", site=site)

    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen_create(self, get_or_create):
        user = User(username="testuser")
        LastActive = mock.Mock(LastActive)
        get_or_create.return_value = (LastActive, True)

        LastActive.objects.seen(user=user)

        get_or_create.assert_called_with(
            user=user, module=settings.last_active_DEFAULT_MODULE, site=Site.objects.get_current()
        )
        self.assertFalse(LastActive.save.called)

    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen_update(self, get_or_create):
        user = User(username="testuser")
        LastActive = mock.Mock(LastActive)
        # force last seen old
        old_time = timezone.now() - datetime.timedelta(seconds=(settings.last_active_INTERVAL * 2))
        LastActive.last_active = old_time
        get_or_create.return_value = (LastActive, False)

        ret = LastActive.objects.seen(user=user)

        get_or_create.assert_called_with(
            user=user, module=settings.last_active_DEFAULT_MODULE, site=Site.objects.get_current()
        )
        self.assertTrue(LastActive.save.called)
        self.assertNotEqual(ret.last_active, old_time)

    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen_update_forced(self, get_or_create):
        user = User(username="testuser")
        LastActive = mock.Mock(LastActive)
        # force last seen old
        old_time = timezone.now()
        LastActive.last_active = old_time
        get_or_create.return_value = (LastActive, False)

        ret = LastActive.objects.seen(user=user, force=True)

        get_or_create.assert_called_with(
            user=user, module=settings.last_active_DEFAULT_MODULE, site=Site.objects.get_current()
        )
        self.assertTrue(LastActive.save.called)
        self.assertNotEqual(ret.last_active, old_time)

    @mock.patch("last_active.models.LastActive.objects.get_or_create", autospec=True)
    def test_seen_found_not_updated(self, get_or_create):
        user = User(username="testuser")
        LastActive = mock.Mock(LastActive)
        # force last seen old
        old_time = timezone.now()
        LastActive.last_active = old_time
        get_or_create.return_value = (LastActive, False)

        ret = LastActive.objects.seen(user=user)

        get_or_create.assert_called_with(
            user=user, module=settings.last_active_DEFAULT_MODULE, site=Site.objects.get_current()
        )
        self.assertFalse(LastActive.save.called)
        self.assertEqual(ret.last_active, old_time)

    def test_when_non_existent(self):
        user = User(username="testuser", pk=1)
        self.assertRaises(LastActive.DoesNotExist, LastActive.objects.when, user)

    @mock.patch("last_active.models.LastActive.objects.filter", autospec=True)
    def test_seen_defaults(self, filter):
        user = User(username="testuser")
        LastActive.objects.when(user=user)

        filter.assert_called_with(user=user)

    @mock.patch("last_active.models.LastActive.objects.filter", autospec=True)
    def test_seen_module(self, filter):
        user = User(username="testuser")
        LastActive.objects.when(user=user, module="mod")

        filter.assert_called_with(user=user, module="mod")

    @mock.patch("last_active.models.LastActive.objects.filter", autospec=True)
    def test_seen_site(self, filter):
        user = User(username="testuser")
        site = Site()
        LastActive.objects.when(user=user, site=site)

        filter.assert_called_with(user=user, site=site)


class TestUserSeen(TestCase):
    def setUp(self):
        [cache.delete(key) for key in list(cache._cache.keys())]

    @mock.patch("last_active.models.LastActive.objects.seen", autospec=True)
    def test_user_seen(self, seen):
        user = User(username="testuser", pk=999)

        user_seen(user)
        site = Site.objects.get_current()
        seen.assert_called_with(user, module=settings.last_active_DEFAULT_MODULE, site=site)

    @mock.patch("last_active.models.LastActive.objects.seen", autospec=True)
    def test_user_seen_no_default(self, seen):
        user = User(username="testuser", pk=1)
        site = Site(pk=2)
        user_seen(user, module="test", site=site)
        seen.assert_called_with(user, module="test", site=site)

    @mock.patch("last_active.models.LastActive.objects.seen", autospec=True)
    def test_user_seen_cached(self, seen):
        user = User(username="testuser", pk=1)
        module = "test_mod"
        cache.set("last_active:%s:%s" % (module, user.pk), time.time())
        user_seen(user, module=module)
        self.assertFalse(seen.called)

    @mock.patch("last_active.models.LastActive.objects.seen", autospec=True)
    def test_user_seen_cache_expired(self, seen):
        user = User(username="testuser", pk=1)
        module = "test_mod"
        cache.set(
            "last_active:%s:%s" % (module, user.pk),
            time.time() - (2 * settings.last_active_INTERVAL),
        )
        user_seen(user, module=module)
        site = Site.objects.get_current()
        seen.assert_called_with(user, module=module, site=site)


class TestClearInterval(TestCase):
    @mock.patch("last_active.models.LastActive.objects.filter", autospec=True)
    @mock.patch("last_active.models.cache", autospec=True)
    def test_clear_interval(self, cache, filter):
        site = Site.objects.get_current()
        user = User(username="testuser", pk=1)
        ls1 = LastActive(user=user, module="mod1", site=site)
        ls2 = LastActive(user=user, module="mod2", site=site)
        filter.return_value = [ls1, ls2]

        clear_interval(user)

        filter.assert_called_with(user=user)
        expected = {"last_active:1:mod1:1": -1, "last_active:1:mod2:1": -1}
        cache.set_many.assert_called_with(expected)

    @mock.patch("last_active.models.LastActive.objects.filter", autospec=True)
    @mock.patch("last_active.models.cache", autospec=True)
    def test_clear_interval_none(self, cache, filter):
        user = User(username="testuser", pk=1)
        filter.return_value = []

        clear_interval(user)

        filter.assert_called_with(user=user)
        self.assertFalse(cache.delete_many.called)

    def test_clear_interval_works(self):
        user = User.objects.create(username="testuser")

        user_seen(user)
        when1 = LastActive.objects.when(user=user)
        clear_interval(user)
        user_seen(user)
        when2 = LastActive.objects.when(user=user)

        self.assertNotEqual(when1, when2)


class TestMiddleware(TestCase):

    middleware = middleware.LastActiveMiddleware()

    @mock.patch("last_active.middleware.user_seen", autospec=True)
    def test_process_request(self, user_seen):
        request = mock.Mock()
        request.user.is_authenticated.return_value = False
        self.middleware.process_request(request)
        self.assertFalse(user_seen.called)

    @mock.patch("last_active.middleware.user_seen", autospec=True)
    def test_process_request_auth(self, user_seen):
        request = mock.Mock()
        request.user.is_authenticated.return_value = True
        self.middleware.process_request(request)
        user_seen.assert_called_with(request.user)



# Generated by Django 3.2.3 on 2021-07-11 11:36

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="LastActive",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("module", models.CharField(default="default", max_length=20)),
                (
                    "last_active",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="sites.site"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-last_active",),
                "unique_together": {("user", "site", "module")},
            },
        ),
    ]



import datetime
import time

from django.contrib.sites.models import Site
from django.core.cache import cache
from django.db import models
from django.utils import timezone

from . import settings


class LastActiveManager(models.Manager):
    """
    Manager for LastActive objects
    Provides 2 utility methods
    """

    def seen(
        self, user, module=settings.LAST_SEEN_DEFAULT_MODULE, site=None, force=False
    ):
        """
        Mask an user last on database seen with optional module and site
        If module not provided uses LAST_SEEN_DEFAULT_MODULE from settings
        If site not provided uses current site
        The last seen object is only updates is LAST_SEEN_INTERVAL seconds
        passed from last update or force=True
        """
        if not site:
            site = Site.objects.get_current()
        args = {
            "user": user,
            "site": site,
            "module": module,
        }
        seen, created = self.get_or_create(**args)
        if created:
            return seen

        # if we get the object, see if we need to update
        limit = timezone.now() - datetime.timedelta(seconds=settings.LAST_SEEN_INTERVAL)
        if seen.last_active < limit or force:
            seen.last_active = timezone.now()
            seen.save()

        return seen

    def when(self, user, module=None, site=None):
        args = {"user": user}
        if module:
            args["module"] = module
        if site:
            args["site"] = site
        return self.filter(**args).latest("last_active").last_active


class LastActive(models.Model):
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    module = models.CharField(default=settings.LAST_SEEN_DEFAULT_MODULE, max_length=20)
    last_active = models.DateTimeField(default=timezone.now)

    objects = LastActiveManager()

    class Meta:
        unique_together = (("user", "site", "module"),)
        ordering = ("-last_active",)

    def __unicode__(self):
        return "%s on %s" % (self.user, self.last_active)


def get_cache_key(site, module, user):
    """
    Get cache database to cache last database write timestamp
    """
    return "last_active:%s:%s:%s" % (site.id, module, user.pk)


def user_seen(user, module=settings.LAST_SEEN_DEFAULT_MODULE, site=None):
    """
    Mask an user last seen on database if LAST_SEEN_INTERVAL seconds
    have passed from last database write.
    Uses optional module and site
    If module not provided uses LAST_SEEN_DEFAULT_MODULE from settings
    If site not provided uses current site
    """
    if not site:
        site = Site.objects.get_current()
    cache_key = get_cache_key(site, module, user)
    # compute limit to update db
    limit = time.time() - settings.LAST_SEEN_INTERVAL
    seen = cache.get(cache_key)
    if not seen or seen < limit:
        # mark the database and the cache, if interval is cleared force
        # database write
        if seen == -1:
            LastActive.objects.seen(user, module=module, site=site, force=True)
        else:
            LastActive.objects.seen(user, module=module, site=site)
        timeout = settings.LAST_SEEN_INTERVAL
        cache.set(cache_key, time.time(), timeout)


def clear_interval(user):
    """
    Clear cached interval from last database write timestamp
    Usefuf if you want to force a database write for an user
    """
    keys = {}
    for last_active in LastActive.objects.filter(user=user):
        cache_key = get_cache_key(last_active.site, last_active.module, user)
        keys[cache_key] = -1

    if keys:
        cache.set_many(keys)



from django.conf import settings

LAST_SEEN_DEFAULT_MODULE = getattr(settings, "LAST_SEEN_DEFAULT_MODULE", "default")
LAST_SEEN_INTERVAL = getattr(settings, "LAST_SEEN_INTERVAL", 60 * 60 * 2)

AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.User")
